#ifndef _BasicGroupBigraph_H
#define _BasicGroupBigraph_H

#include "ProteinIdentifier.h"
#include "ReplicateIndexer.h"
#include "BasicBigraph.h"
#include "Model.h"

#include "Cache.h"

//#define TRUE_BRUTE

#ifdef TRUE_BRUTE
#define NOSEPARATE
#define NOGROUP
#define NOPRUNE
#define NOCACHE
#endif

class Counter
{
 public:
  Counter()
    {
      size = 0;
      start();
    }
  Counter(int s)
    {
      size = s;
      start();
    }
  int size;
  int state;

  void start()
  {
    state = 0;
  }
  bool inRange() const
  {
    return state <= size;
  }
  void advance()
  {
    if ( size == 0 )
      cerr << "Warning: advancing group with size 0" << endl;

    state++;
  }

  friend ostream & operator <<(ostream & os, const Counter & c)
  {
    return os << " Counter(" << c.state << "/" << c.size << ") ";
  }
  
  static void start(Array<Counter> & cA)
  {
    for (int k=0; k<cA.size(); k++)
      cA[k].start();
  }

  static bool inRange(Array<Counter> & cA)
  {
    return cA.back().inRange();
  }

  static void advance(Array<Counter> & cA)
  {
    for (int k=0; k<cA.size(); k++)
      {
	cA[k].advance();
	if ( ! cA[k].inRange() && k != cA.size() - 1 )
	  {
	    // do not reset the last one so that you know when it is not inRange
	    cA[k].start();
	  }
	else
	  {
	    // no need to carry, so you can stop
	    break;
	  }
      }
  }
};


class BasicGroupBigraph : public BasicBigraph
{
  // hack should be protected
 public:
  // protected:

  Array<Counter> originalN;

  Array<Array<string> > groupProtNames;

  Array<double> probabilityR;

  //  Array<double> probabilityE_OverAlphaBeta;

  // protected construction functions
  void groupProteins();
  void trivialGroupProteins();
  void groupProteinsBy(const Array<Set> & groups);
 
  // protected utility functions
  int numberAssociatedProteins(int indexEpsilon) const;
  int numberActiveAssociatedProteins(int indexEpsilon, const Array<Counter> & n) const;
 
  // protected mathematical functions

  double probabilityEEpsilon(const Model & m, int indexEpsilon) const;
  Array<double> probabilityE(const Model & m) const;
  double probabilityEEpsilonGivenActiveAssociatedProteins(const Model & m, int indexEpsilon, int active) const;
  double probabilityNumberAssociatedProteins(const Model & m, int total, int active) const;
  double probabilityEEpsilonGivenN(const Model & m, int indexEpsilon, const Array<Counter> & n) const;

  double likelihoodNGivenD(const Model & m, const Array<Counter> & n) const;
  double logLikelihoodNGivenD(const Model & m, const Array<Counter> & n) const;

  double probabilityN(const Model & m, const Array<Counter> & n) const;
  double probabilityNNu(const Model & m, const Counter & nNu) const;
  double probabilityNGivenD(const Model & m, const Array<Counter> & n) const;

  double logLikelihoodConstant(const Model & m) const;
  double likelihoodConstant(const Model & m) const;

  Array<double> MAPRGivenD(const Model & m);

  Array<double> probabilityRGivenD(const Model & m);
  Array<double> probabilityRGivenN(const Array<Counter> & n);
  double probabilityRRhoGivenN(int indexRho, const Array<Counter> & n) const;

  Array<double> probabilityEGivenD(const Model & m);
  Array<double> eCorrection(const Model & m, const Array<Counter> & n);
  double eCorrectionEpsilon(int indexEpsilon, const Model & m, const Array<Counter> & n);

  Array<double> estimateAlphaGivenD(const Model & m);
  Array<double> alphaEstimateCorrection(const Model & m, const Array<Counter> & n) const;
  double alphaEstimateCorrectionEpsilon(int indexEpsilon, const Model & m, const Array<Counter> & n) const;
  double alphaEstimateCorrectionNuEpsilon(int nu, int indexEpsilon, const Model & m, const Array<Counter> & n, double probRRhoGivenD) const;


  double likelihoodProdTerm(int indexEpsilon, const Model & m, const Array<Counter> & n) const;

  Array<double> probabilityXGivenD(bool xState, const Model & m);
  Array<double> xCorrection(bool xState, const Array<Counter> & n);
  double xCorrectionEpsilon(int indexEpsilon, bool xState, const Array<Counter> & n);

  Array<double> probabilityXAndEGivenD(bool xState, const Model & m);
  Array<double> xAndECorrection(bool xState, const Model & m, const Array<Counter> & n);
  double xAndECorrectionEpsilon(int indexEpsilon, bool xState, const Model & m, const Array<Counter> & n);

  // for unknown alpha, beta
  double probabilityEEpsilonOverAllAlphaBeta(const GridModel & gm, int indexEpsilon) const;
  Array<double> probabilityEOverAllAlphaBeta(const GridModel & gm) const;


  // cache functors
  // note that these will need to be updated if the object is copied
  //  LastCachedMemberFunction<BasicGroupBigraph, double, Model> likelihoodConstantCachedFunctor;
  //  LastCachedMemberFunction<BasicGroupBigraph, Array<double>, Model> probabilityECachedFunctor;
  //  LastCachedMemberFunction<BasicGroupBigraph, Array<double>, GridModel> probabilityEOverAllAlphaBetaCachedFunctor;
  LastCachedMemberFunction<BasicGroupBigraph, double, Model> logLikelihoodConstantCachedFunctor;

 public:
  void refreshCache()
  {
    logLikelihoodConstantCachedFunctor.reset();
  }

  // for partitioning
  double logNumberOfConfigurations() const;

  void getProteinProbs(const Model & m);
  void getMAPProteins(const Model & m);
  void printProteinWeights() const;

  double logLikelihoodAlphaBetaGivenD(const GridModel & gm) const;
  //  double likelihoodAlphaBetaGivenD(const GridModel & gm) const;

  const Array<double> & proteinProbabilities() const
  {
    return probabilityR;
  }
  const Array<Array<string> > & proteinGroupNames() const
  {
    return groupProtNames;
  }

/*  BasicGroupBigraph(): */
/*   likelihoodConstantCachedFunctor( & BasicGroupBigraph::likelihoodConstant, "likelihoodConstant"), probabilityECachedFunctor( & BasicGroupBigraph::probabilityE, "probabilityE"), probabilityEOverAllAlphaBetaCachedFunctor( & BasicGroupBigraph::probabilityEOverAllAlphaBeta, "probabilityEOverAllAlphaBeta"), logLikelihoodConstantCachedFunctor( & BasicGroupBigraph::logLikelihoodConstant, "logLikelihoodConstant") */
/*   { */
/*   } */

 BasicGroupBigraph():
  logLikelihoodConstantCachedFunctor( & BasicGroupBigraph::logLikelihoodConstant, "logLikelihoodConstant")
  {
  }


 void read(istream & fin)
   {
     BasicBigraph::read(fin);
     groupProteins();
   }

 BasicGroupBigraph(const BasicBigraph & rhs) :
 BasicBigraph(rhs),
   //   likelihoodConstantCachedFunctor( & BasicGroupBigraph::likelihoodConstant, "likelihoodConstant"), probabilityECachedFunctor( & BasicGroupBigraph::probabilityE, "probabilityE"),
   //   probabilityEOverAllAlphaBetaCachedFunctor( & BasicGroupBigraph::probabilityEOverAllAlphaBeta, "probabilityEOverAllAlphaBeta"),
   logLikelihoodConstantCachedFunctor( & BasicGroupBigraph::logLikelihoodConstant, "logLikelihoodConstant")
  {
    #ifndef NOGROUP
    groupProteins();
    #else
    trivialGroupProteins();
    #endif

    // these should not vary for alpha and beta
    //    getPeptideProbsOverAllAlphaBeta();

    //    getSumLikelihoodOverAllAlphaBeta();
  }
};

#endif

